import React from 'react';
import { Table } from "react-bootstrap";
import ColorPill from "compendium/components/ColorPill";
import { NrenAndYearDatapoint } from "compendium/Schema";

interface PillTableProps<T extends NrenAndYearDatapoint | { [key: string]: number | string }> {
    columns: string[];
    dataLookup: Map<string, Map<string, Map<number, T>>>;
    circle?: boolean;
    columnLookup?: Map<string, string | null>;
}

function PillTable<T extends NrenAndYearDatapoint | { [key: string]: number | string }>({ columns, dataLookup, circle = false, columnLookup = new Map() }: PillTableProps<T>) {

    const allCategories = Array.from(new Set(Array.from(dataLookup.values()).flatMap(columnData => Array.from(columnData.keys()))));
    const dataColumns = columns.map(column => columnLookup.get(column) || column);

    const allYears = Array.from(new Set(Array.from(dataLookup.values()).flatMap(columnData => Array.from(columnData.values()).flatMap(yearData => Array.from(yearData.keys())))));

    const otherCategories = allCategories.filter(column => {
        const lookupValue = columnLookup.get(column);
        if (lookupValue) {
            return !dataColumns.includes(lookupValue)
        }
        return !dataColumns.includes(column);
    }).map(column => columnLookup.get(column) || column);

    return (<Table className="charging-struct-table" striped bordered >
        <colgroup>
            <col span={1} style={{ width: "12rem" }} />
        </colgroup>
        <thead>
            <tr>
                <th></th>
                {columns.map(column => (
                    <th colSpan={1} key={column}>{column}</th>
                ))}
                {otherCategories.length ? <th>Other</th> : null}
            </tr>

        </thead>
        <tbody>
            {Array.from(dataLookup.entries()).map(([nren, columnData]) => (
                <tr key={nren}>
                    <td>{nren}</td>
                    {dataColumns.map(column => {
                        const dataForYear = columnData.get(column);
                        if (!dataForYear) {
                            return (
                                <td key={column}>
                                </td>
                            );
                        }
                        return (
                            <td key={column}>
                                {allYears.map((year) => {
                                    const data = dataForYear.get(year) || {}
                                    return (
                                        <ColorPill
                                            key={year}
                                            year={year}
                                            active={dataForYear.has(year)}
                                            tooltip={data['tooltip']}
                                            rounded={circle}
                                        />
                                    );
                                })}
                            </td>
                        )
                    })}

                    {!!otherCategories.length && <td key={`${nren}-other`}>
                        {otherCategories.map(otherCategory => {
                            const dataForYear = columnData.get(otherCategory);
                            if (!dataForYear) {
                                // don't return anything if there is no data for an "other" category
                                // because it's freetext and we don't want to display one per year
                                return;
                            }

                            const years = Array.from(Array.from(dataForYear.entries()))
                            return (

                                years.map(([year, data]) => {
                                    return (
                                        <ColorPill
                                            key={year}
                                            year={year}
                                            active
                                            tooltip={data['tooltip'] || otherCategory}
                                            rounded={circle}
                                        />
                                    );
                                })
                            )
                        })}
                    </td>}
                </tr>
            ))}

        </tbody>
    </Table>
    );
}

export default PillTable;
