import React, { useContext } from "react";
import { Link } from "react-router-dom";

import {
    ConnectedProportion,
    ConnectionCarrier,
    ConnectivityGrowth,
    ConnectivityLevel,
    ConnectivityLoad,
    ConnectivityPage
} from "compendium/Schema";
import { createCategoryMatrixLookup } from "compendium/helpers/dataconversion";
import DataPage from "compendium/components/DataPage";
import Filter from "compendium/components/graphing/Filter";
import { CommercialConnectionCategories, CommercialOrgCategories, Sections, UserCategories } from "compendium/helpers/constants";
import { FilterSelectionContext } from "compendium/providers/FilterSelectionProvider";
import ChartContainer from "compendium/components/graphing/ChartContainer";
import { useData } from "compendium/helpers/useData";
import { ScrollableMatrix } from "compendium/components/ScrollableMatrix";
import titles from "compendium/titles";

const displayTitle = {
    [ConnectivityPage.ConnectedProportion]: titles["connected-proportion"],
    [ConnectivityPage.ConnectivityLevel]: titles["connectivity-level"],
    [ConnectivityPage.ConnectionCarrier]: titles["connection-carrier"],
    [ConnectivityPage.ConnectivityLoad]: titles["connectivity-load"],
    [ConnectivityPage.ConnectivityGrowth]: titles["connectivity-growth"],
    [ConnectivityPage.CommercialChargingLevel]: titles["commercial-charging-level"],
    [ConnectivityPage.CommercialConnectivity]: titles["commercial-connectivity"],
}

const description = {
    [ConnectivityPage.ConnectedProportion]: <span>
        European NRENs all have different connectivity remits, as is shown in the table below. The categories of institutions make use of the ISCED 2011 classification system, the UNESCO scheme for International Standard Classification of Education.
        <br />The table shows whether a particular category of institution falls within the connectivity remit of the NREN, the actual number of such institutions connected, the % market share this represents, and the actual number of end users served in the category.
    </span>,
    [ConnectivityPage.ConnectivityLevel]: <span>
        The table below shows the average level of connectivity for each category of institution. The connectivity remit of different NRENs is shown on a different page, and NRENs are asked, at a minimum, to provide information about the typical and highest capacities (in Mbit/s) at which Universities and Research Institutes are connected.
        <br />NRENs are also asked to show proportionally how many institutions are connected at the highest capacity they offer.
    </span>,
    [ConnectivityPage.ConnectionCarrier]: <span>
        The table below shows the different mechanisms employed by NRENs to carry traffic to the different types of users they serve.
        Not all NRENs connect all of the types of institution listed below - details of connectivity remits can be found here: <Link to="/connected-proportion" className="">
            <span>{displayTitle[ConnectivityPage.ConnectedProportion]}</span>
        </Link>
    </span>,
    [ConnectivityPage.ConnectivityLoad]: <span>The table below shows the traffic load in Mbit/s to and from institutions served by NRENs; both the average load, and peak load, when given.
        The types of institutions are broken down using the ISCED 2011 classification system (the UNESCO scheme for International Standard Classification of Education), plus other types.</span>,
    [ConnectivityPage.ConnectivityGrowth]: <span>The table below illustrates the anticipated traffic growth within NREN networks over the next three years.</span>,
    [ConnectivityPage.CommercialChargingLevel]: <span>The table below outlines the typical charging levels for various types of commercial connections.</span>,
    [ConnectivityPage.CommercialConnectivity]: <span>The table below outlines the types of commercial organizations NRENs connect.</span>,
}

const rowFieldMap = {
    [ConnectivityPage.ConnectedProportion]: {
        "Remit cover connectivity": "coverage",
        "Number of institutions connected": "number_connected",
        "Percentage market share of institutions connected": "market_share",
        "Number of users served": "users_served",
    },
    [ConnectivityPage.ConnectivityLevel]: {
        "Typical link speed (Mbit/s):": "typical_speed",
        "Highest speed link (Mbit/s):": "highest_speed",
        "Proportionally how many institutions in this category are connected at the highest capacity? (%):"
            : "highest_speed_proportion",
    },
    [ConnectivityPage.ConnectionCarrier]: {
        "Commercial Provider Backbone": "commercial_provider_backbone",
        "NREN Local Loops": "nren_local_loops",
        "Regional NREN Backbone": "regional_nren_backbone",
        "MAN": "man",
        "Other": "other",

    },
    [ConnectivityPage.ConnectivityLoad]: {
        "Average Load From Institutions (Mbit/s)": "average_load_from_institutions",
        "Average Load To Institutions (Mbit/s)": "average_load_to_institutions",
        "Peak Load To Institution (Mbit/s)": "peak_load_to_institutions",
        "Peak Load From Institution (Mbit/s)": "peak_load_from_institutions"
    },
    [ConnectivityPage.ConnectivityGrowth]: {
        "Percentage growth": "growth"
    },
    [ConnectivityPage.CommercialChargingLevel]: {
        "No charges applied if requested by R&E users": "no_charges_if_r_e_requested",
        "Same charging model as for R&E users": "same_as_r_e_charges",
        "Charges typically higher than for R&E users": "higher_than_r_e_charges",
        "Charges typically lower than for R&E users": "lower_than_r_e_charges"
    },
    [ConnectivityPage.CommercialConnectivity]: {
        "No - but we offer a direct or IX peering": "no_but_direct_peering",
        "No - not eligible for policy reasons": "no_policy",
        "No - financial restrictions (NREN is unable to charge/recover costs)": "no_financial",
        "No - other reason / unsure": "no_other",
        "Yes - National NREN access only": "yes_national_nren",
        "Yes - Including transit to other networks": "yes_incl_other",
        "Yes - only if sponsored by a connected institution": "yes_if_sponsored",
    }
};

interface inputProps {
    page: ConnectivityPage
}

type DataFormat = ConnectedProportion | ConnectivityLevel | ConnectionCarrier | ConnectivityLoad | ConnectivityGrowth

function ConnectedUserPage({ page }: inputProps): React.ReactElement {
    const apiUrl = `/api/connected-${page.toString()}`;

    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, years, nrens } = useData<DataFormat>(apiUrl, setFilterSelection);

    const selectedData = data.filter(data =>
        filterSelection.selectedYears.includes(data.year)
        && filterSelection.selectedNrens.includes(data.nren)
    );

    let isTickIcon = false
    let dataLookup;
    let categoryLookup;

    if (page == ConnectivityPage.CommercialConnectivity) {
        categoryLookup = CommercialOrgCategories;
        isTickIcon = true;
        // case where categories are known in advance, rather than being in the data
        dataLookup = createCategoryMatrixLookup(selectedData, Object.keys(CommercialOrgCategories), undefined);
    } else if (page == ConnectivityPage.CommercialChargingLevel) {
        categoryLookup = CommercialConnectionCategories;
        isTickIcon = true;
        // case where categories are known in advance, rather than being in the data
        dataLookup = createCategoryMatrixLookup(selectedData, Object.keys(CommercialConnectionCategories), undefined);
    } else if (page == ConnectivityPage.ConnectionCarrier) {
        categoryLookup = UserCategories;
        isTickIcon = true;
        dataLookup = createCategoryMatrixLookup(selectedData, ["carry_mechanism"], "user_category");
    } else if (page == ConnectivityPage.ConnectedProportion) {
        categoryLookup = UserCategories;
        dataLookup = createCategoryMatrixLookup(selectedData, Object.values(rowFieldMap[page]), "user_category", false);
    } else {
        categoryLookup = UserCategories;
        dataLookup = createCategoryMatrixLookup(selectedData, Object.values(rowFieldMap[page]), "user_category", false);
    }
    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    // contains title for each row, and the field in the data object to look up for that row
    const rowInfo = rowFieldMap[page];

    const filename = `nren_connected_${page.toString()}`;

    return (
        <DataPage title={displayTitle[page]}
            description={description[page]}
            category={Sections.ConnectedUsers} filter={filterNode}
            data={selectedData} filename={filename}>
            <ChartContainer>
                <ScrollableMatrix
                    dataLookup={dataLookup}
                    rowInfo={rowInfo}
                    isTickIcon={isTickIcon}
                    categoryLookup={categoryLookup}
                />
            </ChartContainer>
        </DataPage>
    );
}
export default ConnectedUserPage;
