import { ReactElement, useContext } from 'react';
import { Line } from 'react-chartjs-2';
import {
    Chart as ChartJS,
    CategoryScale,
    LinearScale,
    PointElement,
    LineElement,
    Title,
    Tooltip,
    Legend,
} from 'chart.js';

import { DarkFibreLease } from "compendium/Schema";
import { createLineChartDataset } from "compendium/helpers/dataconversion";
import DataPage from 'compendium/components/DataPage';
import Filter from "compendium/components/graphing/Filter";
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import ChartContainer from "compendium/components/graphing/ChartContainer";
import { useData } from 'compendium/helpers/useData';
import { getLineChartOptions } from 'compendium/helpers/charthelpers';
import titles from "compendium/titles";

ChartJS.register(
    CategoryScale,
    LinearScale,
    PointElement,
    LineElement,
    Title,
    Tooltip,
    Legend
);

interface Props {
    national?: boolean;
}

function DarkFibreLeasePage({ national }: Props): ReactElement {
    const field = national ? 'fibre_length_in_country' : 'fibre_length_outside_country';
    const validityCheck = (data: DarkFibreLease) => data[field] != null
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, nrens } = useData<DarkFibreLease>('/api/dark-fibre-lease', setFilterSelection, validityCheck);

    const selectedData = data.filter(data =>
        filterSelection.selectedNrens.includes(data.nren) && validityCheck(data) // we only allow filtering nrens for this page
    );
    const dataset = createLineChartDataset(selectedData, field);

    const filterNode = <Filter
        filterOptions={{ availableYears: [], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    const options = getLineChartOptions({ title: 'Kilometres of Leased Dark Fibre', tooltipUnit: 'km', unit: 'km' });

    const description =
        <span>This graph shows the number of Kilometres of dark fibre leased by NRENs {national ? 'within' : 'outside'} their own countries.
            Also included is fibre leased via an IRU (Indefeasible Right of Use), a type of long-term lease of a portion of the capacity of a cable.
            It does not however, include fibre NRENs have installed, and own, themselves.
            The distance is the number of kilometres of the fibre pairs, or if bidirectional traffic on a single fibre, it is treated as a fibre pair.</span>;

    return (
        <DataPage title={national ? titles['dark-fibre-lease'] : titles['dark-fibre-lease-international']}
            description={description}
            category={Sections.Network} filter={filterNode}
            data={selectedData} filename={`dark_fibre_lease_${national ? 'national' : 'international'}`}>
            <ChartContainer>
                <Line data={dataset} options={options} />
            </ChartContainer>
        </DataPage>
    );
}

export default DarkFibreLeasePage;
