import { ReactElement, useContext, useState, useEffect } from "react";
import { useNavigate, Link } from "react-router-dom";
import { Table, Container, Row } from "react-bootstrap";
import * as XLSX from "xlsx";
import SurveySidebar from "compendium/survey/management/SurveySidebar";
import useMatomo from "compendium/matomo/UseMatomo";
import { userContext } from "compendium/providers/UserProvider";
import { fetchSurveys, fetchActiveSurveyYear } from "compendium/survey/api/survey";
import { Survey } from "compendium/survey/api/types";

interface NRENExclusion {
    nren: string;
    year: number;
}

const NrenExclusionsTable = () => {
    const [nrenExclusions, setNrenExclusions] = useState<NRENExclusion[]>([]);
    const [nrens, setNrens] = useState<string[]>([]);

    function fetchNrenExclusions() {
        fetch('/api/nren-exclusions')
            .then(response => response.json())
            .then(data => {
                setNrenExclusions(data);
            })
            .catch(error => {
                console.error('Error fetching NREN exclusions:', error);
            });
    }

    useEffect(() => {
        fetch('/api/nren/list')
            .then(response => response.json())
            .then(data => {
                const nrenNames = data.map(nren => nren.name);
                setNrens(nrenNames);
            })
            .catch(error => {
                console.error('Error fetching NRENs:', error);
            });
    }, []);

    async function addNrenExclusion(nren: string, year: number) {
        const response = await fetch('/api/nren-exclusions', {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json',
            },
            body: JSON.stringify({ nren, year }),
        });

        const json = await response.json();

        if (response.ok) {
            setNrenExclusions(json);
        } else {
            alert(json.message)
        }
        (document.getElementById('excludenren') as HTMLInputElement).value = '';
        (document.getElementById('excludeyear') as HTMLInputElement).value = '';
    }

    async function deleteNrenExclusion(nren: string, year: number) {
        const response = await fetch(`/api/nren-exclusions`, {
            method: 'DELETE',
            headers: {
                'Content-Type': 'application/json',
            },
            body: JSON.stringify({ nren, year }),
        });

        const json = await response.json();

        if (response.ok) {
            setNrenExclusions(json);
        } else {
            alert(json.message)
        }
    }

    useEffect(() => {
        fetchNrenExclusions();
    }, []);

    return (
        <>
            <hr className="fake-divider" />
            <h2>NREN Exclusions</h2>
            <p>Use this table to exclude data for a specific NREN and year.</p>
            <p>If an NREN shows up in this table, they will not be allowed to view/edit or answer their survey.</p>
            <Table striped bordered responsive>
                <thead>
                    <tr>
                        <th>NREN</th>
                        <th>Year</th>
                        <th>Actions</th>
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td>
                            <select id="excludenren">
                                <option value="">Select NREN</option>
                                {nrens.map(nren => (
                                    <option key={nren} value={nren}>{nren}</option>
                                ))}
                            </select>
                        </td>
                        <td>
                            <input id="excludeyear" type="year" name="year" />
                        </td>
                        <td>
                            <button className="btn btn-primary" onClick={() => {
                                const nren = (document.getElementById('excludenren') as HTMLInputElement).value;
                                const year = parseInt((document.getElementById('excludeyear') as HTMLInputElement).value);
                                if (!nren || !year) {
                                    alert('Please enter an NREN and year');
                                    return;
                                }
                                addNrenExclusion(nren, year);
                            }}>Add Exclusion</button>
                        </td>
                    </tr>
                    {nrenExclusions.map(exclusion => (
                        <tr key={exclusion.nren + exclusion.year}>
                            <td>{exclusion.nren}</td>
                            <td>{exclusion.year}</td>
                            <td>
                                <button className="btn btn-danger" onClick={
                                    () => deleteNrenExclusion(exclusion.nren, exclusion.year)
                                }>Delete</button>
                            </td>
                        </tr>
                    ))}
                </tbody>
            </Table>
            <hr className="fake-divider" />
        </>

    );
}

const SurveyTable = () => {

    const [survey, setSurvey] = useState<Survey>();

    useEffect(() => {
        fetchSurveys().then((surveyList) => {
            // only show the latest survey
            setSurvey(surveyList[0]);
        });
    }, []);

    return (<Table striped bordered responsive>
        <thead>
            <tr>
                <th>(N)REN</th>
                <th>Link</th>
                <th>Survey Status</th>
            </tr>
        </thead>
        <tbody>
            {survey && survey.responses.map(response => (
                <tr key={response.nren.id}>
                    <td>{response.nren.name}</td>
                    <td>
                        <Link to={`/survey/response/${survey.year}/${response.nren.name}`}><span>Navigate to survey</span></Link>
                    </td>
                    <td>{response.status}</td>
                </tr>
            ))}

        </tbody>
    </Table>)
}

function Landing(): ReactElement {

    const { trackPageView } = useMatomo();

    const { user } = useContext(userContext);
    const navigate = useNavigate();

    const loggedIn = !!user.id;
    const hasNren = loggedIn ? !!user.nrens.length : false;
    const activeNren = hasNren ? user.nrens[0] : '';
    const isAdmin = loggedIn ? user.permissions.admin : false;
    const isObserver = loggedIn ? user.role === 'observer' : false;

    const [activeSurveyYear, setActiveSurveyYear] = useState<string | null>(null);

    useEffect(() => {
        const fetchData = async () => {
            const year = await fetchActiveSurveyYear();
            setActiveSurveyYear(year);
        };

        fetchData();

        trackPageView({ documentTitle: "GEANT Survey Landing Page" });
    }, [trackPageView]);

    const moveToSurvey = () => {
        try {
            navigate(`/survey/response/${activeSurveyYear}/${activeNren}`);
            return <li>Redirecting to survey...</li>; // Or a loading indicator
        } catch (error) {
            console.error("Error navigating:", error);
            return null; // Or display an error message
        }
    };

    function applyCustomFormat(ws: XLSX.WorkSheet, columnName: string, format: string) {
        /* Get the worksheet range */
        const range = XLSX.utils.decode_range(ws['!ref'] ?? '');
        let columnIndex = -1;
        for (let i = range.s.c; i <= range.e.c; i++) {
            const cellRef = XLSX.utils.encode_cell({ r: range.s.r, c: i });
            const cell = ws[cellRef];
            if (cell && typeof cell.v === 'string' && cell.v === columnName) {
                columnIndex = i;
                break;
            }
        }
        /* If the column index was not found, return */
        if (columnIndex === -1) {
            console.error(`Column '${columnName}' not found.`);
            return;
        }
        for (let i = range.s.r + 1; i <= range.e.r; ++i) {
            const ref = XLSX.utils.encode_cell({ r: i, c: columnIndex });
            if (!ws[ref]) continue;
            if (ws[ref].t !== 'n') continue;
            ws[ref].z = format;
        }
    }

    function convertToExcel(jsonData: { name: string, data, meta }[]): Blob {
        const wb = XLSX.utils.book_new();
        jsonData.forEach(sheet => {
            const ws = XLSX.utils.json_to_sheet(sheet.data);
            if (sheet.meta) {
                applyCustomFormat(ws, sheet.meta.columnName, sheet.meta.format);
            }
            XLSX.utils.book_append_sheet(wb, ws, sheet.name);

        })
        const wbout = XLSX.write(wb, { bookType: 'xlsx', type: 'binary' });

        const buffer = new ArrayBuffer(wbout.length);
        const view = new Uint8Array(buffer);
        for (let i = 0; i < wbout.length; i++) {
            // Convert each character of the binary workbook string to an 8-bit integer and store in the Uint8Array 'view' for blob creation.
            view[i] = wbout.charCodeAt(i) & 0xFF;
        }

        return new Blob([buffer], { type: 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet;charset=UTF-8' });
    }

    function fetchDataAndConvertToExcel() {
        const apiEndpoint = '/api/data-download';

        fetch(apiEndpoint)
            .then(response => {
                if (!response.ok) {
                    throw new Error('Network response was not ok');
                }
                return response.json();
            })
            .then(data => {
                // Call convertToExcel function with the retrieved data
                const excelBlob = convertToExcel(data);

                // Create a download link and trigger a click event to download the file
                const downloadLink = document.createElement('a');
                downloadLink.href = URL.createObjectURL(excelBlob);
                downloadLink.download = 'data.xlsx';
                document.body.appendChild(downloadLink);
                downloadLink.click();
                document.body.removeChild(downloadLink);
            })
            .catch(error => {
                console.error('Error fetching data:', error);
                alert('An error occurred while creating the data download Excel file.');
            });
    }

    return (
        <>
            {isAdmin && <SurveySidebar />}
            <Container className="py-5 grey-container">
                <Row>
                    <div className="center-text">
                        <h1 className="geant-header">THE GÉANT COMPENDIUM OF NRENS SURVEY</h1>

                        <div className="wordwrap pt-4" style={{ maxWidth: '75rem' }}>
                            <p style={{ textAlign: "left" }}>
                                Hello,
                                <br />Welcome to the GÉANT Compendium Survey. (N)REN Compendium administrators can login via Single Sign On (SSO) <a href="/login">here</a>, which will complete their registration to fill in the latest Compendium survey.
                                This will send a notification to the Compendium administration team and they will assign you to your (N)REN.
                                <br />Once this step has been completed, you will receive an email from the administration team. We aim to get back to you the same working day, but sometimes may take a little longer.
                                <br />If you are not sure whether you are a Compendium Administrator for your (N)REN, please contact your GÉANT Partner Relations relationship manager.
                                <br />Thank you.
                            </p>

                            <span>Current registration status:</span>
                            <br />
                            <br />
                            {isAdmin ? <ul>
                                <li><span>You are logged in as a Compendium Administrator</span></li>
                                <li><span>Click <Link to="/survey/admin/surveys">here</Link> to access the survey management page.</span></li>
                                <li><span>Click <Link to="/survey/admin/users">here</Link> to access the user management page.</span></li>
                                <li><span>Click <a href="#" onClick={fetchDataAndConvertToExcel}>here</a> to do the full data download.</span></li>
                            </ul> : <ul>
                                {activeSurveyYear && !isAdmin && !isObserver && hasNren && moveToSurvey()}
                                {loggedIn ? <li><span>You are logged in</span></li> : <li><span>You are not logged in</span></li>}
                                {loggedIn && !isObserver && !hasNren && <li><span>Your access to the survey has not yet been approved</span></li>}
                                {loggedIn && !isObserver && !hasNren && <li><span>Once you have been approved, you will immediately be directed to the relevant survey upon visiting this page</span></li>}
                                {loggedIn && isObserver && <li><span>You have read-only access to the following surveys:</span></li>}
                            </ul>}
                            {isAdmin && <NrenExclusionsTable />}
                            {loggedIn && isObserver && <SurveyTable />}
                        </div>
                    </div>
                </Row>
            </Container >
        </>
    );
}

export default Landing;
