import logging
import openpyxl

from pathlib import Path

from compendium_v2.conversion import mapping
from compendium_v2.db.presentation_model_enums import CarryMechanism, ConnectivityCoverage, MonitoringMethod, \
    UserCategory, FeeType, YesNoPlanned
from compendium_v2.environment import setup_logging
from compendium_v2.publishers.helpers import map_nren

setup_logging()

logger = logging.getLogger(__name__)

EXCEL_FILE_ORGANISATION = openpyxl.load_workbook(str(Path(__file__).resolve(
).parent.parent / "resources" / "2021_Organisation_DataSeries.xlsx"), data_only=True, read_only=True)
EXCEL_FILE_USERS = openpyxl.load_workbook(str(Path(__file__).resolve(
).parent.parent / "resources" / "2022_Connected_Users_DataSeries.xlsx"), data_only=True, read_only=True)
EXCEL_FILE_NETWORKS = openpyxl.load_workbook(str(Path(__file__).resolve(
).parent.parent / "resources" / "2022_Networks_DataSeries.xlsx"), data_only=True, read_only=True)
EXCEL_FILE_NREN_SERVICES = openpyxl.load_workbook(str(Path(__file__).resolve(
).parent.parent / "resources" / "NREN-Services-prefills_2023_Recovered.xlsx"), data_only=True, read_only=True)


def fetch_budget_excel_data():
    ws = EXCEL_FILE_ORGANISATION["1. Budget"]
    for row in ws.iter_rows(min_row=14, max_row=58, min_col=1, max_col=8):
        nren = map_nren(row[1].value)
        for i, year in zip(range(3, 8), reversed(range(2015, 2022))):
            budget = row[i].value
            if budget is not None:
                budget = round(budget / 1000000, 2)
                assert isinstance(nren, str) and isinstance(budget, float) and isinstance(year, int)
                yield nren.upper(), budget, year


def fetch_funding_excel_data():
    ws = EXCEL_FILE_ORGANISATION["2. Income Sources"]
    ws2 = EXCEL_FILE_ORGANISATION["Income Sources table"]

    def hard_number_convert(s, source_name, nren, year):
        if s is None:
            return float(0)
        try:
            return float(s)
        except ValueError:
            return float(0)

    # iterate over the rows in the worksheet
    def create_points_for_year_until_2017(start_row, end_row, year, col_nren, col_start):
        for row in ws.iter_rows(min_row=start_row, max_row=end_row, min_col=col_nren, max_col=col_start+5):
            offset = col_start - col_nren
            # extract the data from the row
            nren = map_nren(row[0].value)
            client_institution = row[offset].value
            commercial = row[offset+1].value
            geant_subsidy = row[offset+2].value
            gov_public_bodies = row[offset+3].value
            other_european_funding = row[offset+4].value
            other = row[offset+5].value

            client_institution = hard_number_convert(client_institution, "client institution", nren, year)
            commercial = hard_number_convert(commercial, "commercial", nren, year)
            geant_subsidy = hard_number_convert(geant_subsidy, "geant subsidy", nren, year)
            gov_public_bodies = hard_number_convert(gov_public_bodies, "gov/public_bodies", nren, year)
            other_european_funding = hard_number_convert(other_european_funding, "other european funding", nren, year)
            other = hard_number_convert(other, "other", nren, year)

            european_funding = geant_subsidy + other_european_funding

            # process the data (e.g. save to database)
            if nren is not None:
                if nren.upper() == 'HEANET':
                    other = other + commercial
                    commercial = 0
                yield (nren.upper(), year, client_institution, european_funding, gov_public_bodies, commercial, other)

    def create_points_for_year_from_2018(sheet, start_row, end_row, year, nren_col, col_start):
        for row in sheet.iter_rows(min_row=start_row, max_row=end_row, min_col=nren_col, max_col=col_start+4):
            offset = col_start - nren_col
            # extract the data from the row
            nren = map_nren(row[0].value)
            client_institution = row[offset].value
            european_funding = row[offset+1].value
            gov_public_bodies = row[offset+2].value
            commercial = row[offset+3].value
            other = row[offset+4].value

            client_institution = hard_number_convert(client_institution, "client institution", nren, year)
            european_funding = hard_number_convert(european_funding, "european funding", nren, year)
            gov_public_bodies = hard_number_convert(gov_public_bodies, "gov/public_bodies", nren, year)
            commercial = hard_number_convert(commercial, "commercial", nren, year)
            other = hard_number_convert(other, "other", nren, year)

            # process the data (e.g. save to database)
            if nren is not None:
                if nren.upper() == 'HEANET':
                    other = other + commercial
                    commercial = 0
                yield (nren.upper(), year, client_institution, european_funding, gov_public_bodies, commercial, other)

    # For 2016
    yield from create_points_for_year_until_2017(8, 51, 2016, 43, 45)

    # For 2017
    yield from create_points_for_year_until_2017(8, 51, 2017, 32, 35)

    # For 2018
    yield from create_points_for_year_from_2018(ws, 8, 51, 2018, 21, 24)

    # For 2019
    yield from create_points_for_year_from_2018(ws, 8, 51, 2019, 12, 15)

    # For 2020
    yield from create_points_for_year_from_2018(ws, 8, 51, 2020, 3, 6)

    # For 2021
    yield from create_points_for_year_from_2018(ws2, 8, 51, 2021, 11, 12)


def fetch_charging_structure_excel_data():
    ws = EXCEL_FILE_ORGANISATION["3. Charging mechanism"]

    # iterate over the rows in the worksheet
    def create_points_for_2021(start_row, end_row, year, col_start):
        for row in range(start_row, end_row):
            # extract the data from the row
            nren = map_nren(ws.cell(row=row, column=col_start).value)
            charging_structure = ws.cell(row=row, column=col_start + 1).value
            if charging_structure is not None:
                if "do not charge" in charging_structure:
                    charging_structure = FeeType.no_charge
                elif "combination" in charging_structure:
                    charging_structure = FeeType.combination
                elif "flat" in charging_structure:
                    charging_structure = FeeType.flat_fee
                elif "usage-based" in charging_structure:
                    charging_structure = FeeType.usage_based_fee
                elif "Other" in charging_structure:
                    charging_structure = FeeType.other
                else:
                    charging_structure = None

                yield nren.upper(), year, charging_structure

    def create_points_for_2019(start_row, end_row, year, col_start):
        for row in range(start_row, end_row):
            # extract the data from the row
            nren = map_nren(ws.cell(row=row, column=col_start).value)
            charging_structure = ws.cell(row=row, column=col_start + 1).value
            structure_2 = ws.cell(row=row, column=col_start + 2).value
            if charging_structure is not None:
                if structure_2 not in [None, '']:
                    charging_structure = FeeType.other
                elif "do not charge" in charging_structure:
                    charging_structure = FeeType.no_charge
                elif "combination" in charging_structure:
                    charging_structure = FeeType.combination
                elif "flat" in charging_structure:
                    charging_structure = FeeType.flat_fee
                elif "usage-based" in charging_structure:
                    charging_structure = FeeType.usage_based_fee
                elif "Other" in charging_structure:
                    charging_structure = FeeType.other
                else:
                    charging_structure = None

                yield nren.upper(), year, charging_structure

    # For 2021
    yield from create_points_for_2021(3, 46, 2021, 2)

    # For 2019
    yield from create_points_for_2019(3, 46, 2019, 6)


def fetch_staffing_excel_data():
    ws = EXCEL_FILE_ORGANISATION["4. Staff"]

    start_row = 18
    end_row = 61

    def convert_number(value, nren, year, description):
        try:
            return float(value)
        except (TypeError, ValueError):
            return 0

    def create_points_for_year(year, nren_column, start_column):
        offset = start_column - nren_column
        for row in ws.iter_rows(min_row=start_row, max_row=end_row, min_col=nren_column, max_col=start_column + 1):

            # extract the data from the row
            nren = map_nren(row[0].value)
            if not nren:
                continue
            permanent = row[offset].value
            permanent = convert_number(permanent, nren, year, "permanent ftes")
            subcontracted = row[offset+1].value
            subcontracted = convert_number(subcontracted, nren, year, "subcontractor ftes")
            if permanent + subcontracted > 0:
                yield nren.upper(), year, permanent, subcontracted

    yield from create_points_for_year(2021, 2, 5)
    yield from create_points_for_year(2020, 13, 16)
    yield from create_points_for_year(2019, 23, 26)
    yield from create_points_for_year(2018, 33, 36)
    yield from create_points_for_year(2017, 43, 46)
    yield from create_points_for_year(2016, 53, 55)


def fetch_staff_function_excel_data():
    ws = EXCEL_FILE_ORGANISATION["5. Staff by Function"]

    start_row = 14
    end_row = 58

    def convert_number(value, nren, year, description):
        try:
            return float(value)
        except (TypeError, ValueError):
            return 0

    def create_points_for_year_until_2019(year, nren_column, start_column):
        offset = abs(start_column - nren_column)
        for row in ws.iter_rows(min_row=start_row, max_row=end_row, min_col=nren_column, max_col=start_column + 5):
            # extract the data from the row
            nren = map_nren(row[0].value)
            if nren is None:
                continue
            admin = convert_number(row[offset].value, nren, year, "admin and finance ftes")
            communication = convert_number(row[offset+1].value, nren, year, "communication ftes")
            infosec = convert_number(row[offset+2].value, nren, year, "infosec ftes")
            it = convert_number(row[offset+3].value, nren, year, "it and software dev ftes")
            noc = convert_number(row[offset+4].value, nren, year, "NOC and engineering ftes")
            others = convert_number(row[offset+5].value, nren, year, "other ftes")

            technical = infosec + it + noc
            non_technical = admin + communication + others

            if technical + non_technical > 0:
                yield nren.upper(), year, technical, non_technical

    def create_points_for_year(year, nren_column, start_column):
        offset = abs(start_column - nren_column)
        for row in ws.iter_rows(min_row=start_row, max_row=end_row, min_col=nren_column, max_col=start_column + 1):
            nren = map_nren(row[0].value)
            if nren is None:
                continue
            technical = convert_number(row[offset].value, nren, year, "technical ftes")
            non_technical = convert_number(row[offset+1].value, nren, year, "non-technical ftes")
            if technical + non_technical > 0:
                yield nren.upper(), year, technical, non_technical

    # For 2017
    yield from create_points_for_year_until_2019(2017, 41, 43)

    # For 2018
    yield from create_points_for_year_until_2019(2018, 31, 33)

    # For 2019
    yield from create_points_for_year_until_2019(2019, 20, 22)

    # For 2020
    yield from create_points_for_year(2020, 12, 14)

    # For 2021
    yield from create_points_for_year(2021, 3, 5)


def fetch_ecproject_excel_data():
    ws = EXCEL_FILE_ORGANISATION["7. EC Projects"]

    start_row = 6

    def create_points_for_year(year, start_column, end_row):
        for row in ws.iter_rows(min_row=start_row, max_row=end_row, min_col=start_column, max_col=start_column + 1):
            # extract the data from the row
            nren = map_nren(row[0].value)
            if nren is None:
                continue
            project = row[1].value
            if project is not None:
                yield nren.upper(), year, project

    yield from create_points_for_year(2017, 13, 165)

    yield from create_points_for_year(2018, 10, 165)

    yield from create_points_for_year(2019, 7, 165)

    yield from create_points_for_year(2020, 4, 180)

    yield from create_points_for_year(2021, 1, 173)


def fetch_organization_excel_data():
    ws = EXCEL_FILE_ORGANISATION["Organization"]

    # iterate over the rows in the worksheet
    for row in range(5, 48):
        # extract the data from the row
        nren = map_nren(ws.cell(row=row, column=2).value)
        parent_org = ws.cell(row=row, column=4).value

        if parent_org not in [None, 'NA', 'N/A']:
            yield nren.upper(), 2021, parent_org


def fetch_traffic_excel_data():
    ws = EXCEL_FILE_NETWORKS["Estimated_Traffic TByte"]

    rows = list(ws.rows)

    def convert_number(value, nren, year, description):
        if value is None or value == '--' or value == 'No data':
            return 0
        try:
            return float(value)
        except (TypeError, ValueError):
            return 0

    def create_points_for_year(year, start_column):
        for i in range(6, 49):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())

            from_external = convert_number(rows[i][start_column + 1].value, nren_name, year, 'from_external')
            to_external = convert_number(rows[i][start_column + 2].value, nren_name, year, 'to_external')
            from_customer = convert_number(rows[i][start_column + 3].value, nren_name, year, 'from_customer')
            to_customer = convert_number(rows[i][start_column + 4].value, nren_name, year, 'to_customer')
            if from_external == 0 and to_external == 0 and from_customer == 0 and to_customer == 0:
                continue

            yield nren_name, year, from_external, to_external, from_customer, to_customer

    yield from create_points_for_year(2016, 38)
    yield from create_points_for_year(2017, 32)
    yield from create_points_for_year(2018, 26)
    yield from create_points_for_year(2019, 20)
    yield from create_points_for_year(2020, 14)
    yield from create_points_for_year(2021, 8)
    yield from create_points_for_year(2022, 2)


def fetch_nren_services_excel_data():
    ws = EXCEL_FILE_NREN_SERVICES["Sheet1"]
    rows = list(ws.rows)

    titles = rows[0]

    nren_service_data_columns = {}

    def normalize_nren_name(n: str) -> str:
        n = n.split(' ')[0].upper()
        return {'KIFÜ': 'KIFU', 'ANAS': 'AZSCIENCENET', 'PIONIER': 'PSNC'}.get(n, n)

    for i in range(0, 131):
        if titles[i].value:
            name = normalize_nren_name(titles[i].value)
            nren_service_data_columns[name] = i

    for nren_name, start_column in nren_service_data_columns.items():
        for row_index in range(2, 61):
            row = rows[row_index]
            service_name = row[0].value
            if row[start_column].value and row[start_column].value.upper() == 'YES':
                product_name = ''
                additional_information = ''
                if row[start_column + 1].value:
                    product_name = row[start_column + 1].value
                if row[start_column + 2].value:
                    additional_information = row[start_column + 2].value

                if service_name not in mapping.SERVICES_MAPPING:
                    continue

                service_category, service_name_key = mapping.SERVICES_MAPPING[service_name].split(':')
                service_category = mapping.SERVICE_CATEGORY_MAPPING[service_category]

                yield {
                    'nren_name': nren_name,
                    'service_category': service_category,
                    'service_name_key': service_name_key,
                    'year': 2022,
                    'product_name': product_name.strip(),
                    'additional_information': additional_information.strip(),
                    'official_description': '',
                }


def get_category(excel_cat):
    if not excel_cat:
        return None
    if "hospital" in excel_cat.lower():
        return UserCategory.hospitals
    if "universit" in excel_cat.lower():
        return UserCategory.universities
    if "research ins" in excel_cat.lower():
        return UserCategory.institutes
    if "further" in excel_cat.lower() or "fe" == excel_cat.lower():
        return UserCategory.further_education
    if "inter" in excel_cat.lower():
        return UserCategory.iros
    if "cultural" in excel_cat.lower() or "librar" in excel_cat.lower():
        return UserCategory.cultural
    if "primary" in excel_cat.lower():
        return UserCategory.primary_schools
    if "secondary" in excel_cat.lower():
        return UserCategory.secondary_schools
    if "govern" in excel_cat.lower():
        return UserCategory.government
    if "profit" in excel_cat.lower():
        return UserCategory.for_profit_orgs


def fetch_remit_excel_data():
    ws = EXCEL_FILE_USERS["Connectivity Remit"]

    def get_remit(excel_remit):
        if not excel_remit:
            return None
        if "including transit" in excel_remit.lower():
            return ConnectivityCoverage.yes_incl_other
        if "national nren" in excel_remit.lower():
            return ConnectivityCoverage.yes_national_nren
        if "some circ" in excel_remit.lower():
            return ConnectivityCoverage.sometimes
        if "policy reas" in excel_remit.lower():
            return ConnectivityCoverage.no_policy
        if "financial" in excel_remit.lower():
            return ConnectivityCoverage.no_financial
        if "other reason" in excel_remit.lower():
            return ConnectivityCoverage.no_other
        if "unsure" in excel_remit.lower():
            return ConnectivityCoverage.unsure

    result = {}

    def create_points_for_year(year, start_column, cut=False):
        category_row = [c.value for c in list(ws.iter_rows(
            min_row=8, max_row=8, min_col=start_column+2, max_col=start_column + 21 + (2 if cut else 0)))[0]]
        categories = [get_category(cat) for cat in category_row[::2]]
        if cut:
            # remove the 3rd value, which is the nren name column
            categories = categories[:2] + categories[3:]

        for row in ws.iter_rows(min_row=9, max_row=52, min_col=start_column+1,
                                max_col=start_column + 21 + (2 if cut else 0)):

            nren_name = row[0].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            every_second_value = [c.value for c in row][::2]
            # remove the first value, which is the nren name
            every_second_value = every_second_value[1:]
            if cut:
                # remove the 3rd value, which is another nren name
                every_second_value = every_second_value[:2] + every_second_value[3:]

            remit = [get_remit(val) for val in every_second_value]

            values = dict(zip(categories, remit))

            for category, remit in values.items():
                if remit:
                    result[(nren_name, year, category)] = remit

    create_points_for_year(2022, 3)
    create_points_for_year(2021, 26, cut=True)
    create_points_for_year(2020, 50)
    create_points_for_year(2019, 72)
    return result


def fetch_nr_connected_excel_data():
    ws = EXCEL_FILE_USERS["Connected Institutions"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(5, 48):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 11):
                category = get_category(rows[4][c].value)
                nr_connected = int(rows[i][c].value) if rows[i][c].value else None
                if category and nr_connected:
                    result[(nren_name, year, category)] = nr_connected

    create_points_for_year(2019, 39)
    create_points_for_year(2020, 27)
    create_points_for_year(2021, 14)
    create_points_for_year(2022, 2)
    return result


def fetch_market_share_excel_data():
    ws = EXCEL_FILE_USERS["Table Market Share"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(8, 51):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 11):
                category = get_category(rows[7][c].value)
                percentage_connected = float(rows[i][c].value) if rows[i][c].value else None
                if category and percentage_connected:
                    result[(nren_name, year, category)] = percentage_connected

    create_points_for_year(2017, 64)
    create_points_for_year(2018, 52)
    create_points_for_year(2019, 40)
    create_points_for_year(2020, 28)
    create_points_for_year(2021, 16)
    create_points_for_year(2022, 3)
    return result


def fetch_users_served_excel_data():
    ws = EXCEL_FILE_USERS["Users"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(4, 47):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 11):
                category = get_category(rows[3][c].value)
                users_connected = int(rows[i][c].value) if rows[i][c].value else None
                if category and users_connected:
                    result[(nren_name, year, category)] = users_connected

    create_points_for_year(2019, 40)
    create_points_for_year(2020, 28)
    create_points_for_year(2021, 14)
    create_points_for_year(2022, 2)
    return result


def fetch_typical_speed_excel_data():
    ws = EXCEL_FILE_USERS["Table _Typical IP Link capacity"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(33, 76):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 11):
                category = get_category(rows[32][c].value)
                typical_speed = int(rows[i][c].value) if rows[i][c].value else None
                if category and typical_speed:
                    result[(nren_name, year, category)] = typical_speed

    create_points_for_year(2017, 75)
    create_points_for_year(2018, 50)
    create_points_for_year(2019, 38)
    create_points_for_year(2020, 26)
    create_points_for_year(2021, 14)
    create_points_for_year(2022, 2)
    return result


def fetch_highest_speed_excel_data():
    ws = EXCEL_FILE_USERS["Table _Highest IP Link capacity"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(33, 76):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 11):
                category = get_category(rows[32][c].value)
                highest_speed = int(rows[i][c].value) if rows[i][c].value else None
                if category and highest_speed:
                    result[(nren_name, year, category)] = highest_speed

    create_points_for_year(2017, 64)
    create_points_for_year(2018, 51)
    create_points_for_year(2019, 38)
    create_points_for_year(2020, 26)
    create_points_for_year(2021, 14)
    create_points_for_year(2022, 2)
    return result


def fetch_highest_speed_proportion_excel_data():
    ws = EXCEL_FILE_USERS["Aver High cap conn Share"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(5, 48):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 11):
                category = get_category(rows[4][c].value)
                highest_speed = float(rows[i][c].value) if rows[i][c].value else None
                if category and highest_speed:
                    result[(nren_name, year, category)] = highest_speed

    create_points_for_year(2020, 27)
    create_points_for_year(2021, 14)
    create_points_for_year(2022, 2)
    return result


def fetch_carriers_excel_data():
    ws = EXCEL_FILE_USERS["Traffic carriers"]
    rows = list(ws.rows)

    def get_carrier(excel_carrier):
        if not excel_carrier:
            return None
        if "comme" in excel_carrier.lower():
            return CarryMechanism.commercial_provider_backbone
        if "man" in excel_carrier.lower():
            return CarryMechanism.man
        if "local loop" in excel_carrier.lower():
            return CarryMechanism.nren_local_loops
        if "other" in excel_carrier.lower():
            return CarryMechanism.other
        if "regional" in excel_carrier.lower():
            return CarryMechanism.regional_nren_backbone

    result = {}

    def create_points_for_year(year, start_column):
        for i in range(3, 46):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 11):
                category = get_category(rows[2][c].value)
                carrier = get_carrier(rows[i][c].value)
                if category and carrier:
                    result[(nren_name, year, category)] = carrier

    create_points_for_year(2019, 40)
    create_points_for_year(2020, 27)
    create_points_for_year(2021, 14)
    create_points_for_year(2022, 2)
    return result


def fetch_growth_excel_data():
    ws = EXCEL_FILE_NETWORKS["Table Traffic Growth % "]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(5, 46):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 11):
                category = get_category(rows[4][c].value)
                growth = float(rows[i][c].value) if rows[i][c].value else None
                if category and growth:
                    result[(nren_name, year, category)] = growth

    create_points_for_year(2019, 40)
    create_points_for_year(2020, 26)
    create_points_for_year(2021, 14)
    create_points_for_year(2022, 2)
    return result


def fetch_average_traffic_excel_data():
    ws = EXCEL_FILE_USERS["Average Traffic"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(5, 48):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 21, 2):
                category = get_category(rows[3][c].value)
                from_inst = int(rows[i][c].value) if rows[i][c].value else None
                to_inst = int(rows[i][c+1].value) if rows[i][c+1].value else None
                if category and (from_inst or to_inst):
                    result[(nren_name, year, category)] = (from_inst, to_inst)

    create_points_for_year(2019, 68)
    create_points_for_year(2020, 46)
    create_points_for_year(2021, 24)
    create_points_for_year(2022, 2)
    return result


def fetch_peak_traffic_excel_data():
    ws = EXCEL_FILE_USERS["Peak traffic"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(6, 49):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            for c in range(start_column + 1, start_column + 21, 2):
                category = get_category(rows[4][c].value)
                from_inst = int(rows[i][c].value) if rows[i][c].value else None
                to_inst = int(rows[i][c+1].value) if rows[i][c+1].value else None
                if category and (from_inst or to_inst):
                    result[(nren_name, year, category)] = (from_inst, to_inst)

    create_points_for_year(2019, 70)
    create_points_for_year(2020, 47)
    create_points_for_year(2021, 24)
    create_points_for_year(2022, 2)
    return result


def fetch_remote_campuses_excel_data():
    ws = EXCEL_FILE_USERS["Foreign Campuses"]
    rows = list(ws.rows)

    def create_points_for_year(year, start_column):
        for i in range(5, 48):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            have_remote = rows[i][start_column + 1].value
            connectivity = rows[i][start_column + 2].value
            country = rows[i][start_column + 3].value
            connected_to_r_e = rows[i][start_column + 4].value
            if have_remote and have_remote.upper() == "YES":
                connectivity = connectivity.upper() == "YES" if connectivity else False
                connected_to_r_e = connected_to_r_e not in [None, "-", "Not connected.", "We do not know"]
                country = country or ""
                yield nren_name, year, connectivity, country, connected_to_r_e

    yield from create_points_for_year(2019, 22)
    yield from create_points_for_year(2020, 16)
    yield from create_points_for_year(2021, 10)
    yield from create_points_for_year(2022, 4)


def fetch_dark_fibre_iru_excel_data():
    ws = EXCEL_FILE_NETWORKS["Dark Fibre"]
    rows = list(ws.rows)

    def parse_int(excel_value):
        if excel_value is None or excel_value == "":
            return None
        return int(str(excel_value).replace("-", "").replace(" ", "").replace(".", ""))

    def create_points_for_year(year, start_column):
        for i in range(10, 53):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            s = start_column
            iru = ""
            if year > 2019:
                s += 1
                iru = rows[i][s].value
            elif parse_int(rows[i][s + 1].value) is not None:
                iru = "Yes" if parse_int(rows[i][s + 1].value) else "No"

            if iru:
                length_in_country = parse_int(rows[i][s + 1].value)
                length_out_country = parse_int(rows[i][s + 3].value)
                iru = iru == "Yes"
                yield nren_name, year, iru, length_in_country, length_out_country

    yield from create_points_for_year(2016, 42)
    yield from create_points_for_year(2017, 36)
    yield from create_points_for_year(2018, 30)
    yield from create_points_for_year(2019, 24)
    yield from create_points_for_year(2020, 17)
    yield from create_points_for_year(2021, 10)
    yield from create_points_for_year(2022, 3)


def fetch_dark_fibre_installed_excel_data():
    ws = EXCEL_FILE_NETWORKS["Dark Fibre"]
    rows = list(ws.rows)

    def parse_int(excel_value):
        if excel_value is None or excel_value == "":
            return None
        return int(str(excel_value).replace("-", "").replace(" ", "").replace(".", ""))

    def create_points_for_year(year, start_column):
        for i in range(10, 53):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            s = start_column
            if year > 2019:
                s += 1
            installed_length = parse_int(rows[i][s + 2].value)
            if installed_length is not None:
                installed = bool(installed_length)
                yield nren_name, year, installed, installed_length

    yield from create_points_for_year(2016, 42)
    yield from create_points_for_year(2017, 36)
    yield from create_points_for_year(2018, 30)
    yield from create_points_for_year(2019, 24)
    yield from create_points_for_year(2020, 17)
    yield from create_points_for_year(2021, 10)
    yield from create_points_for_year(2022, 3)


def fetch_iru_duration_excel_data():
    ws = EXCEL_FILE_NETWORKS["IRU duration"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(3, 46):
            nren_name = rows[i][start_column].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            years = rows[i][start_column + 1].value
            if isinstance(years, str):
                if 'just lease' in years:  # mapping invalid answer
                    years = "0"
            if not years:
                continue
            years = str(years).split(" ")[0].split("+")[0].split("-")[0]
            if not years:
                continue
            try:
                years = int(years)
            except ValueError:
                continue
            result[(nren_name, year)] = years

    create_points_for_year(2019, 10)
    create_points_for_year(2020, 7)
    create_points_for_year(2021, 4)
    create_points_for_year(2022, 1)
    return result


def fetch_passive_monitoring_excel_data():
    ws = EXCEL_FILE_NETWORKS["Traffic monitoring"]
    rows = list(ws.rows)

    def create_points_for_year(year, start_column):
        for i in range(6, 48):
            nren_name = rows[i][1].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            monitoring = rows[i][start_column].value
            method = rows[i][start_column + 1].value
            if monitoring:
                monitoring = monitoring == "Yes"
                method = {
                    "SPAN ports": MonitoringMethod.span_ports,
                    "Passive optical TAPS": MonitoringMethod.taps,
                    "Both": MonitoringMethod.both,
                    None: None
                }[method]
                yield nren_name, year, monitoring, method

    yield from create_points_for_year(2021, 4)
    yield from create_points_for_year(2022, 2)


def fetch_largest_link_capacity_excel_data():
    ws = EXCEL_FILE_NETWORKS["Largest IP Trunk capacity"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(5, 47):
            nren_name = rows[i][5].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            largest_capacity = rows[i][start_column].value
            if largest_capacity:
                result[(nren_name, year)] = int(largest_capacity)

    create_points_for_year(2016, 12)
    create_points_for_year(2017, 11)
    create_points_for_year(2018, 10)
    create_points_for_year(2019, 9)
    create_points_for_year(2020, 8)
    create_points_for_year(2021, 7)
    create_points_for_year(2022, 6)
    return result


def fetch_typical_backbone_capacity_excel_data():
    ws = EXCEL_FILE_NETWORKS["Typical IP Trunk capacity"]
    rows = list(ws.rows)
    result = {}

    def create_points_for_year(year, start_column):
        for i in range(5, 47):
            nren_name = rows[i][4].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            typical_capacity = rows[i][start_column].value
            if typical_capacity:
                result[(nren_name, year)] = int(typical_capacity)

    create_points_for_year(2016, 11)
    create_points_for_year(2017, 10)
    create_points_for_year(2018, 9)
    create_points_for_year(2019, 8)
    create_points_for_year(2020, 7)
    create_points_for_year(2021, 6)
    create_points_for_year(2022, 5)
    return result


def fetch_non_r_e_peers_excel_data():
    ws = EXCEL_FILE_NETWORKS["Peering-Non R& Network"]
    rows = list(ws.rows)

    def create_points_for_year(year, start_column):
        for i in range(5, 48):
            nren_name = rows[i][2].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            nr_peers = rows[i][start_column].value
            if nr_peers:
                yield nren_name, year, int(nr_peers)

    yield from create_points_for_year(2016, 10)
    yield from create_points_for_year(2017, 9)
    yield from create_points_for_year(2018, 8)
    yield from create_points_for_year(2019, 7)
    yield from create_points_for_year(2020, 6)
    yield from create_points_for_year(2021, 5)
    yield from create_points_for_year(2022, 3)


def fetch_ops_automation_excel_data():
    ws = EXCEL_FILE_NETWORKS["Automation"]
    rows = list(ws.rows)

    def create_points_for_year(year, start_column):
        for i in range(5, 48):
            nren_name = rows[i][1].value
            if not nren_name:
                continue
            nren_name = map_nren(nren_name.upper())
            automation = rows[i][start_column].value
            specifics = rows[i][start_column + 1].value or ""
            if automation:
                automation = YesNoPlanned[automation.lower()]
                yield nren_name, year, automation, specifics

    yield from create_points_for_year(2021, 5)
    yield from create_points_for_year(2022, 3)
