"""
survey_publisher_legacy
=========================

This script is used to import data from the legacy datasources into the new survey system.
Legacy data is used to generate a new survey for each NREN, for each year that we have data for.
The generated surveys are based on the 2024 survey.

"""
import click
from itertools import chain
from collections import defaultdict
from typing import Dict, Any

import compendium_v2
from compendium_v2.db import db
from compendium_v2.config import load
from compendium_v2.db.presentation_models import NREN
from compendium_v2.survey_db import model as survey_model
from compendium_v2.db.survey_models import SurveyResponse, Survey, SurveyStatus, ResponseStatus, SurveyNotes
from compendium_v2.publishers.legacy_publisher.survey_publisher_legacy_db import fetch_data as fetch_data_db
from compendium_v2.publishers.legacy_publisher.survey_publisher_legacy_excel import fetch_data as fetch_data_excel

VALID_KEYS_2024_SURVEY = {
    "abbreviation_english",
    "abbreviation_national_languages",
    "alienwave_internal",
    "alienwave_services",
    "alienwave_services_number",
    "audit_specifics",
    "audits",
    "budget",
    "business_continuity_plans",
    "business_continuity_plans_specifics",
    "central_procurement_amount",
    "central_software_procurement",
    "certificate_service",
    "certificate_service-Comment",
    "charging_mechanism",
    "city",
    "commercial_charging_levels",
    "commercial_organizations",
    "commodity_vs_r_e",
    "connected_sites_lists",
    "connected_users_comments",
    "connectivity_level",
    "connectivity_proportions",
    "corporate_strategy",
    "corporate_strategy_url",
    "country",
    "crisis_exercises",
    "crisis_management_procedure",
    "dark_fibre_lease",
    "dark_fibre_lease_duration",
    "dark_fibre_lease_kilometers_inside_country",
    "dark_fibre_lease_kilometers_outside_country",
    "dark_fibre_nren",
    "dark_fibre_nren_kilometers_inside_country",
    "ec_project_names",
    "ec_projects",
    "email_address",
    "external_connections",
    "fibre_light",
    "fibre_light-Comment",
    "formal_service_management_framework",
    "full_name_english",
    "full_name_national_languages",
    "income_sources",
    "max_capacity",
    "monitoring_tools",
    "monitoring_tools-Comment",
    "netflow_vendors",
    "network_automation",
    "network_automation_tasks",
    "network_comments",
    "network_map_urls",
    "network_weather",
    "network_weather_url",
    "nfv",
    "nfv_types",
    "nfv_types-Comment",
    "non_r_and_e_peers",
    "operational_process_automation",
    "operational_process_automation_tools",
    "organization_comments",
    "parent_organization",
    "parent_organization_name",
    "passive_monitoring",
    "passive_monitoring_tech",
    "pert_team",
    "phone_number",
    "policies",
    "policy_comments",
    "postal_code",
    "remote_campuses",
    "remote_campuses_specifics",
    "security_controls",
    "security_controls-Comment",
    "service_comments",
    "service_level_targets",
    "service_matrix",
    "service_portfolio_eosc_portal",
    "services_collaboration",
    "services_hosting",
    "services_identity",
    "services_isp",
    "services_multimedia",
    "services_network",
    "services_on_eosc_portal_list",
    "services_professional",
    "services_security",
    "siem_soc_vendor",
    "siem_soc_vendor-Comment",
    "staff_employment_type",
    "staff_roles",
    "street_name_and_number",
    "suborganization_details",
    "suborganizations",
    "traffic_carriers",
    "traffic_estimate",
    "traffic_growth",
    "traffic_load",
    "traffic_statistics",
    "traffic_statistics_urls",
    "typical_capacity",
    "website"
}


def insert_survey_data(survey_2024: Survey, nren: NREN, year: int, answer: Dict[str, Any]):
    # we're basing the generated survey on the 2024 survey, so we need to make sure that exists
    # before we insert the responses.
    survey = db.session.query(Survey).filter(Survey.year == year).first()
    if not survey:
        survey = Survey(
            year=year,
            survey=survey_2024.survey,
            status=SurveyStatus.published
        )
        db.session.add(survey)
    else:
        # only really necessary for the 2022 survey, which is already in the database.
        survey.survey = survey_2024.survey
        survey.status = SurveyStatus.published
        db.session.add(survey)
    response = db.session.query(SurveyResponse).filter(SurveyResponse.survey_year ==
                                                       year, SurveyResponse.nren_id == nren.id).first()
    # add some default values for the survey
    answer['page'] = 0
    answer['verification_status'] = {}
    if response:
        response.answers = answer
        db.session.add(response)
    else:
        response = SurveyResponse(
            survey=survey,
            survey_year=year,
            nren_id=nren.id,
            nren=nren,
            answers=answer,
            status=ResponseStatus.completed
        )

        response_notes = SurveyNotes(
            survey=response,
            survey_year=year,
            nren_id=nren.id,
            notes="This survey has been imported by the legacy survey importer. Please review the data for accuracy.",
        )

        db.session.add(response)
        db.session.add(response_notes)
    db.session.commit()


def delete_surveys():
    db.session.query(SurveyNotes).filter(SurveyNotes.survey_year >= 2016, SurveyNotes.survey_year <= 2021).delete()
    db.session.query(SurveyResponse).filter(SurveyResponse.survey_year >=
                                            2016, SurveyResponse.survey_year <= 2021).delete()
    db.session.query(Survey).filter(Survey.year >= 2016, Survey.year <= 2021).delete()
    db.session.commit()


def ensure_string_tree(tree: Dict[str, Any]):
    # this function converts all object values to strings (from int, float, etc)
    # if we encounter lists or dicts, we recurse into them.

    for key, value in list(tree.items()):
        if isinstance(value, dict):
            ensure_string_tree(value)
        elif isinstance(value, list):
            for i, val in list(enumerate(value)):
                if isinstance(val, dict):
                    ensure_string_tree(val)
                else:
                    value[i] = str(val)
        else:
            if value is None:
                del tree[key]
            else:
                tree[key] = str(value)
    return tree


def _cli(app):

    with app.app_context():
        nren_map = defaultdict(lambda: defaultdict(lambda: {'data': {}}))

        survey_2022 = db.session.query(SurveyResponse).filter(SurveyResponse.survey_year == 2022).all()

        for resp in survey_2022:
            # for the 2022 survey it's already initialized with the data, but it's not always valid.
            # initialize the data with the 2022 survey data, and then overwrite it with the import, keeping old data.
            nren_map[resp.nren][resp.survey_year]['data'] = resp.answers['data']

        for data_key, nren, nren_id, year, value in chain(fetch_data_excel(), fetch_data_db()):
            if data_key not in VALID_KEYS_2024_SURVEY:
                print(f'Invalid data key: {data_key} for NREN: {nren} ({nren_id}) in year {year}')

            # overwrite the data if we have a new value
            nren_map[nren][year]['data'][data_key] = value

        # use this to gauge quality of a survey year:
        # answers = [len(d['data']) for yearmap in nren_map.values() for year, d in yearmap.items() if year == 2018]
        # sum(answers) / len(answers)
        delete_surveys()
        survey_2024 = db.session.query(Survey).filter(Survey.year == 2024).first()
        for nren, years in nren_map.items():
            for year, data in years.items():
                if year < 2016 or year > 2022:
                    # data before 2016 is very sparse, so don't move it over.
                    # from 2023 onwards it's all handled by the new survey system.
                    continue
                data = ensure_string_tree(data)
                if not data or not data['data']:
                    continue
                insert_survey_data(survey_2024, nren, year, data)


@click.command()
@click.option('--config', type=click.STRING, default='config.json')
def cli(config):
    app_config = load(open(config, 'r'))

    app_config['SQLALCHEMY_BINDS'] = {survey_model.SURVEY_DB_BIND: app_config['SURVEY_DATABASE_URI']}

    app = compendium_v2._create_app_with_db(app_config)
    print("survey-publisher-legacy starting")
    _cli(app)


if __name__ == "__main__":
    cli()
