import logging
from flask import Blueprint, jsonify, request
from sqlalchemy import func
from compendium_v2.auth.session_management import admin_required
from compendium_v2.db import db
from compendium_v2.helpers import get_data
from compendium_v2.routes import common
from compendium_v2.routes.survey import routes as survey
from compendium_v2.routes.user import routes as user_routes
from compendium_v2.routes.response import routes as response_routes
from compendium_v2.routes.data_download import routes as data_download
from compendium_v2.db.presentation_models import (
    NRENExclusion,
    BudgetEntry, FundingSource, ChargingStructure, NrenStaff, ParentOrganization, SubOrganization,
    ECProject, Policy, TrafficVolume, InstitutionURLs, CentralProcurement, ServiceManagement,
    ServiceUserTypes, EOSCListings, Standards, CrisisExercises, SecurityControls, ConnectedProportion,
    ConnectivityLevel, ConnectionCarrier, ConnectivityLoad, ConnectivityGrowth, CommercialConnectivity,
    CommercialChargingLevel, RemoteCampuses, DarkFibreLease, DarkFibreInstalled, FibreLight, NetworkMapUrls,
    MonitoringTools, PassiveMonitoring, TrafficStatistics, SiemVendors, CertificateProviders, WeatherMap,
    PertTeam, AlienWave, Capacity, ExternalConnections, NonREPeers, TrafficRatio, OpsAutomation,
    NetworkFunctionVirtualisation, NetworkAutomation, NREN, NRENService
)


routes = Blueprint('compendium-v2-api', __name__)
routes.register_blueprint(survey, url_prefix='/survey')
routes.register_blueprint(user_routes, url_prefix='/user')
routes.register_blueprint(response_routes, url_prefix='/response')
routes.register_blueprint(data_download, url_prefix='/data-download')
logger = logging.getLogger(__name__)

models = [
    (AlienWave, '/alien-wave', 'ALIEN_WAVE_RESPONSE_SCHEMA'),
    (BudgetEntry, '/budget', 'BUDGET_RESPONSE_SCHEMA'),
    (Capacity, '/capacity', 'CAPACITY_RESPONSE_SCHEMA'),
    (CentralProcurement, '/central-procurement', 'CENTRAL_PROCUREMENT_RESPONSE_SCHEMA'),
    (CertificateProviders, '/certificate-providers', 'CERTIFICATE_PROVIDER_RESPONSE_SCHEMA'),
    (ChargingStructure, '/charging', 'CHARGING_STRUCTURE_RESPONSE_SCHEMA'),
    (CommercialChargingLevel, '/connected-charging', 'COMMERCIAL_CHARGING_LEVEL_RESPONSE_SCHEMA'),
    (CommercialConnectivity, '/connected-commercial', 'COMMERCIAL_CONNECTIVITY_RESPONSE_SCHEMA'),
    (ConnectedProportion, '/connected-proportion', 'CONNECTED_PROPORTION_RESPONSE_SCHEMA'),
    (ConnectionCarrier, '/connected-carrier', 'CONNECTION_CARRIER_RESPONSE_SCHEMA'),
    (ConnectivityGrowth, '/connected-growth', 'CONNECTIVITY_GROWTH_RESPONSE_SCHEMA'),
    (ConnectivityLevel, '/connected-level', 'CONNECTIVITY_LEVEL_RESPONSE_SCHEMA'),
    (ConnectivityLoad, '/connected-load', 'CONNECTIVITY_LOAD_RESPONSE_SCHEMA'),
    (CrisisExercises, '/crisis-exercises', 'CRISIS_EXERCISES_RESPONSE_SCHEMA'),
    (DarkFibreInstalled, '/dark-fibre-installed', 'DARK_FIBRE_INSTALLED_RESPONSE_SCHEMA'),
    (DarkFibreLease, '/dark-fibre-lease', 'DARK_FIBRE_LEASE_RESPONSE_SCHEMA'),
    (ECProject, '/ec-project', 'EC_PROJECTS_RESPONSE_SCHEMA'),
    (EOSCListings, '/eosc-listings', 'EOSC_LISTINGS_RESPONSE_SCHEMA'),
    (ExternalConnections, '/external-connections', 'EXTERNAL_CONNECTIONS_RESPONSE_SCHEMA'),
    (FibreLight, '/fibre-light', 'FIBRE_LIGHT_RESPONSE_SCHEMA'),
    (FundingSource, '/funding', 'FUNDING_RESPONSE_SCHEMA'),
    (InstitutionURLs, '/institution-urls', 'INSTITUTION_URLS_RESPONSE_SCHEMA'),
    (MonitoringTools, '/monitoring-tools', 'MONITORING_TOOLS_RESPONSE_SCHEMA'),
    (NetworkAutomation, '/network-automation', 'NETWORK_AUTOMATION_RESPONSE_SCHEMA'),
    (NetworkFunctionVirtualisation, '/nfv', 'NETWORK_FUNCTION_VIRTUALISATION_RESPONSE_SCHEMA'),
    (NetworkMapUrls, '/network-map-urls', 'NETWORK_MAP_URLS_RESPONSE_SCHEMA'),
    (NonREPeers, '/non-re-peers', 'NON_R_AND_E_PEERS_RESPONSE_SCHEMA'),
    (NREN, '/nren/list', 'NREN_RESPONSE_SCHEMA'),
    (NRENService, '/nren-services', 'NREN_SERVICES_RESPONSE_SCHEMA'),
    (NrenStaff, '/staff', 'NREN_STAFF_RESPONSE_SCHEMA'),
    (OpsAutomation, '/ops-automation', 'OPS_AUTOMATION_RESPONSE_SCHEMA'),
    (ParentOrganization, '/parent-organizations', 'ORGANIZATION_RESPONSE_SCHEMA'),
    (PassiveMonitoring, '/passive-monitoring', 'PASSIVE_MONITORING_RESPONSE_SCHEMA'),
    (PertTeam, '/pert-team', 'PERT_TEAM_RESPONSE_SCHEMA'),
    (Policy, '/policy', 'POLICY_RESPONSE_SCHEMA'),
    (RemoteCampuses, '/remote-campuses', 'REMOTE_CAMPUSES_RESPONSE_SCHEMA'),
    (SecurityControls, '/security-controls', 'SECURITY_CONTROLS_RESPONSE_SCHEMA'),
    (ServiceManagement, '/service-management', 'SERVICE_MANAGEMENT_RESPONSE_SCHEMA'),
    (ServiceUserTypes, '/services-offered', 'SERVICE_USER_TYPES_RESPONSE_SCHEMA'),
    (SiemVendors, '/siem-vendors', 'SIEM_VENDORS_RESPONSE_SCHEMA'),
    (Standards, '/standards', 'STANDARDS_RESPONSE_SCHEMA'),
    (SubOrganization, '/sub-organizations', 'ORGANIZATION_RESPONSE_SCHEMA'),
    (TrafficRatio, '/traffic-ratio', 'TRAFFIC_RATIO_RESPONSE_SCHEMA'),
    (TrafficStatistics, '/traffic-stats', 'TRAFFIC_STATISTICS_RESPONSE_SCHEMA'),
    (TrafficVolume, '/traffic-volume', 'TRAFFIC_VOLUME_RESPONSE_SCHEMA'),
    (WeatherMap, '/weather-map', 'WEATHER_MAP_RESPONSE_SCHEMA')
]


@routes.route('/nren-exclusions', methods=['GET', 'POST', 'DELETE'])
@admin_required
@common.require_accepts_json
def nren_exclusions():

    if request.method == 'POST':
        data = request.get_json()
        nren_name = data.get('nren')
        year = data.get('year')

        nren = db.session.query(NREN).filter(func.lower(NREN.name) == nren_name.lower()).first()
        if not nren:
            return jsonify({'message': f'NREN {nren_name} not found'}), 400

        exists = next((exclusion for exclusion in nren.exclusions if exclusion.year == year), None)
        if exists:
            return jsonify({'message': f'Exclusion for {nren_name} in {year} already exists'}), 400

        exclusion = NRENExclusion(nren=nren, year=year)
        db.session.add(exclusion)
        db.session.commit()

    elif request.method == 'DELETE':
        data = request.get_json()
        nren_name = data.get('nren')
        year = data.get('year')

        nren = db.session.query(NREN).filter(func.lower(NREN.name) == nren_name.lower()).first()
        if not nren:
            return jsonify({'message': f'NREN {nren_name} not found'}), 400

        exclusion = next((exclusion for exclusion in nren.exclusions if exclusion.year == year), None)
        if not exclusion:
            return jsonify({'message': f'Exclusion for {nren_name} in {year} not found'}), 400

        db.session.delete(exclusion)
        db.session.commit()

    exclusions = db.session.query(NRENExclusion).all()
    return jsonify([exclusion.to_dict() for exclusion in exclusions])


def create_view_func(model):
    return lambda: jsonify(get_data(model))


for model, endpoint, schema in models:
    view_func = create_view_func(model)
    view_func.__name__ = f'{model.__name__.lower()}_view'
    view_func.__doc__ = f"""
    Handler for /api{endpoint} requests. Returns data for the {model.__name__} presentation model.

    {model.__doc__ or ''}

    Response will be formatted as:

    .. asjson::
        compendium_v2.routes.schemas.{schema}
    """

    routes.route(f'{endpoint}', methods=['GET'])(common.require_accepts_json(view_func))


@routes.after_request
def after_request(resp):
    return common.after_request(resp)
